/*
versio    1.0
yritys    Tietomato Oy
tekij    Tapio Lehtimki
			
asiakas   MikroBitti-lehti 

Javan sieluokasta "Thread" periytetty filosofiluokka.

Historia
04.10.2002  T~L Luonti	
*/


import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.sql.*;

public class Philosopher extends Thread
{ 
	public static final int STATE_NOTHING     = 0;	
	public static final int STATE_THINK       = 1;
	public static final int STATE_HUNGRY      = 2;
	public static final int STATE_GOT1STICK   = 3;		
	public static final int STATE_WAIT2STICK  = 4;	
	public static final int STATE_EAT         = 5;
	public static final int STATE_STARVED     = 6;
	public static final int STATE_INTERRUPTED = 99;
	
	private final int N_SHOW_ONESTICK_TIME = 500;	
	private final int N_CATCHSTICK_TIME    = 100;	
	
	private int mnState;
	private PhListener moListener;
	private int mnId;
	private int mnAppetite;
	private int mnCatchStick;
	private int mnLunchHour;	
	private int miLeftStick;
	private int miRightStick;
	
	/* 
	********************************************************************************
	Konstruktori 
	********************************************************************************
	*/		
	public Philosopher(int nId, int iLeftStick, int iRightStick,
		PhListener oListener, int nAppetite, int nStarving, int nLunchHour) 
	{
		mnId = nId;
		mnState = STATE_NOTHING;
		setAppetite(nAppetite);
		setStarving(nStarving);
		setLunchHour(nLunchHour);
		miLeftStick = iLeftStick;
		miRightStick = iRightStick;		
		moListener = oListener;	
	}
	
	/* 
	********************************************************************************
	Sikeen kynnistys
	Paikoitellen esiintyv komento "yield()" on tarpeen joissain kyttjrjestelmiss.
	Kyseisen hihkaisun on tarkoitus varmistaa, ett kaistaa j muillekin sikeille. 
	Normaalisti	sikeiden tulisi pelata rinnakkain ilman yield-komentoakin, mutta nin 
	ei kuitenkaan aina ky. Eip yieldin kytt toisaalta mitn haittaakaan.
	********************************************************************************
	*/		
	public void run()
	{
		long i;
		boolean b;
		
		/* Ennen kuin filosofi pstetn pytn viivstetn ukkoa 0-4 sekuntia, muuten
			deadlock iskee liian herksti, kun joka ij ryhtyy tapailemaan sympuikkoja
			samalla kellonlymll */
		try
		{sleep(mnId * 1000);}
		catch (Exception e)
		{}										
		
		/* Kun filosofisie kynnistetn, se pysyy hengiss, kunnes helvetti jtyy, 
		Alko kuivuu tai kone kaatuu. Ellei filosofi ehdi nnty nlkn ennen sit. */
		until_hell_freezes:				
			while(!isInterrupted())
			{
				yield();
				/*  ENSIMMISEN PUIKON METSSTYS				
					Nlkinen filosofi jahtaa sympuikkoa oikealta puoleltaan.
					Hn yritt kopata puikon lyhkisin vliajoin yh uudelleen */	
				synchronized(moListener)
				{setState(STATE_HUNGRY);}
				
				i = 0;
				while (!isInterrupted())
				{
					yield();
					if (isInterrupted())
					{break until_hell_freezes;}
					
					/* Lukitaan kuuntelija puikon kiinniottoyrityksen ajaksi */
					synchronized(moListener)
					{
						b = moListener.getStick(miRightStick);
						if (b)
						{setState(STATE_GOT1STICK);}
					}
					/*  Filosofi sai ensimmisen puikon kiinni. Kuuntelijaa ei pidet
						lukittuna sin aikana, kun filosofi hehkuttaa ainoaa puikkoaan.
						Kuuntelija lukitaan hetkeksi vasta tilaviesti lhetettess. */
					if (b)
					{
						try
						{sleep(N_SHOW_ONESTICK_TIME);}	
						catch(InterruptedException e)
						{break until_hell_freezes;}
						
						synchronized(moListener)
						{setState(STATE_WAIT2STICK);}								
						break;
					}
					/*  Filosofi ei saanut ensimmist puikkoa kiinni. Mikn ei ole lukossa,
						kun filosofi huilaa hetken ennen uutta yrityst. */					
					if (!b)
					{
						try
						{sleep(N_CATCHSTICK_TIME);}
						catch(InterruptedException e)
						{break until_hell_freezes;}
					}
					
					/*  Mikli ekan tikun tavoittelulle on asetettu raja, noudatetaan sit. 
						Muuussa tapauksessa yritetn ikuisesti. */
					if (mnCatchStick > 0)
					{
						i = i + 1;
						if (i > mnCatchStick)
						{break;}
					}
				}
				
				/*  NLKKUOLEMA
					Filosofilla ei saanut kiinni ensimmist puikkoa, joten hn nntyy nlkn
					ja masennukseen. Kuuntelija lukkoon ja tilaviesti eteenpin. */		
				if (mnState == STATE_HUNGRY)
				{
					synchronized(moListener)
					{setState(STATE_STARVED);}
					break until_hell_freezes;
				}
				
				/*  TOISEN PUIKON SAALISTUS
					Filosofilla on jo hallussaan ensimminen puikko - nyt pit viel napata toinen
					tikku lautasen vasemmalta puolelta. En ajattelijaa ei voi nujertaa nlk eik
					masennus.*/		
				i = 0;
				while (!isInterrupted())
				{
					yield();					
					/*  Jos filosofi saa toisenkin puikon ktseens, alkaa ppern popsinta.
						Lukitaan kuuntelija poimintayrityksen ajaksi, mutta vapautetaan se ennen
						kuin sie nukutetaan ruokalevolle. 
						Univaiheen aikana muut sikeet saavat mellastaa vapaasti. */
					synchronized(moListener)
					{
						b = moListener.getStick(miLeftStick); 
						if (b)
						{setState(STATE_EAT);}
					}
					if (b)
					{
						/* Pistetn sie unosille siksi aikaa, kun filosofi mssilee. */
						try
						{sleep(mnLunchHour);}
						catch(InterruptedException e)
						{break until_hell_freezes;}
						
						/*  Sytyn filosofi laskee puikot pydlle ja aloittaa ajattelutyn. 
							Lukitaan kuuntelija siksi aikaa, ett saadaan puikot laskettua 
							pydlle muiden filosofien kytettvksi. */	
						
						synchronized(moListener)
						{													
							moListener.leaveStick(miRightStick);
							moListener.leaveStick(miLeftStick);	
							break;							
						}
					}
					
					/* Tnne pdytn, jos filosofi ei pssyt viel symn. Kuuntelijaa ei ole
						lukittu eik lukita nytkn, kun sie vet hetken henke, ennen kuin yritt
						uudelleen psy lihapatojen reen */
					if (!b)
					{
						try
						{sleep(N_CATCHSTICK_TIME);}
						catch(InterruptedException e)
						{break until_hell_freezes;}
					}
					i = i + 1;
				}
				
				/*  AJATTELUVAIHE "COGITO ERGO SUM"	
					Ruokailun lomassa filosofi tietty fundeerailee, kunnes nlk ylltt.
					Kuuntelija lukitaan tilaviestin lhettmisen ajaksi. */	
				synchronized(moListener)
				{setState(STATE_THINK);}
				
				try
				{sleep(mnAppetite);}
				catch (InterruptedException e)
				{break until_hell_freezes;}								
			}
		
		frozen_hell:
			if (isInterrupted())
			{setState(STATE_INTERRUPTED);}
	}
	
	/* 
	********************************************************************************
	OMINAISUUKSIEN LUKUMETODIT
	********************************************************************************
	*/
	
	public int getState()
	{return mnState;}	
	
	/* 
	********************************************************************************
	OMINAISUUKSIEN ASETUSMETODIT
	********************************************************************************
	
	********************************************************************************
	Tilan voi muuttaa vain olio itse
	********************************************************************************
	*/		
	private void setState(int value)
	{
		/* Tilamuutoksista viestitn mys kuuntelijalle */
		mnState = value;
		moListener.ZarahustraSpeaks(mnId, mnState);	
		yield();		
	}
	
	/* 
	********************************************************************************
	Ruokahalu
	********************************************************************************
	*/	
	public void setAppetite(int value)
	{mnAppetite = 20000 / value;}
	
	/* 
	********************************************************************************
	Nlnsietokyky on suoraan riippuvainen siit, kuinka monta kertaa filosofi
	jaksaa yritt poimia tikun pydlt. Jos ukko ei saa sympuikkoa tss
	asetetun yritysmrn aikana, hn nntyy nlkn. 
	
	Nolla-arvo merkitsee rajatonta sietokyky - filosofi poimii puikkoa pydlt 
	vaikka ikuisesti.
	********************************************************************************
	*/		
	public void setStarving(int value)
	{
		/*  Paras mahdollinen nlnsietokyky muutetaan arvoksi nolla, jonka
			filosofi tulkitsee rajoittamattomaksi sietokyvyksi */
		if (value == Constants.N_STARVING_MAX) 
		{value = Constants.N_NEVER_STARVING;}
		mnCatchStick = value;
	}	
	
	/* 
	********************************************************************************
	Ruokailun kesto eli kauanko filosofista nytetn suu ammollaan ahmivaa kuvaa. 
	Ruokailun ajaksi filosofi varaa kaksi sympuikkoa.
	********************************************************************************
	*/	
	public void setLunchHour(int value)
	{mnLunchHour = value * 100;}	
}

